from dataclasses import dataclass
import numpy
import numpy.typing


@dataclass
class DotMeasurement():
	'''
	Representation of a dot measuring static dipoles.

	Parameters
	----------
	v : float
		The voltage measured at the dot.
	r : numpy.ndarray
		The position of the dot.
	'''
	v: float
	r: numpy.ndarray

	def __post_init__(self) -> None:
		self.r = numpy.array(self.r)

	def v_for_point(self, pt: numpy.ndarray) -> float:
		'''
		Returns the voltage for a static dipole represented as a phase space point.

		Parameters
		----------
		pt: A length 6 numpy array representing a dipole, as (px, py, pz, sx, sy, sz).

		Returns
		----------
		Returns the voltage from that dipole at the point.
		'''
		p = pt[0:3]  # hardcoded here because chances
		s = pt[3:6]  # are we'll only ever work in 3d.

		diff = self.r - s
		return p.dot(diff) / (numpy.linalg.norm(diff)**3)

	def cost(self, pts: numpy.ndarray) -> float:
		# 6 because dipole in 3d has 6 degrees of freedom.
		pt_length = 6
		# creates numpy.ndarrays in groups of pt_length.
		# Will throw problems for irregular points, but that's okay for now.
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]
		return sum(self.v_for_point(pt) for pt in chunked_pts) - self.v

	def jac_pt(self, pt: numpy.ndarray) -> numpy.ndarray:
		p = pt[0:3]  # hardcoded here because chances
		s = pt[3:6]  # are we'll only ever work in 3d.

		diff = self.r - s

		p_divs = diff / (numpy.linalg.norm(diff)**3)

		r_divs = -p / (numpy.linalg.norm(diff)**3) + 3 * p.dot(diff) * diff / (numpy.linalg.norm(diff)**5)

		return numpy.append(p_divs, r_divs)

	def jac(self, pts: numpy.ndarray) -> numpy.ndarray:
		# 6 because dipole in 3d has 6 degrees of freedom.
		pt_length = 6
		# creates numpy.ndarrays in groups of pt_length.
		# Will throw problems for irregular points, but that's okay for now.
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]

		return numpy.append([], [self.jac_pt(pt) for pt in chunked_pts])
