from typing import Callable, Sequence
import numpy
import scipy.optimize
from pathfinder.model.static.dot import DotMeasurement


class DotDipoleModel():
	'''
	Model of n static dipoles with a collection of voltage measurements
	at dots at different positions.

	Parameters
	----------
	dots : Sequence[Dot]
		A collection of dots representing a series of measured voltages.
	n: int
		The number of dipoles to assume.
	'''
	def __init__(self, dots: Sequence[DotMeasurement], n: int) -> None:
		self.dots = dots
		self.m = len(dots)
		self.n = n

	def __repr__(self) -> str:
		return f'DotDipoleModel({repr(list(self.dots))}, {self.n})'

	def costs(self) -> Callable[[numpy.ndarray], numpy.ndarray]:
		def costs_to_return(pt: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([dot.cost(pt) for dot in self.dots])

		return costs_to_return

	def jac(self) -> Callable[[numpy.ndarray], numpy.ndarray]:
		def jac_to_return(pts: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([dot.jac(pts) for dot in self.dots])

		return jac_to_return

	def sol(self, initial_dipole=(0.1, 0.1, 0.1), initial_position=(.1, .1, .1), use_root=True):
		initial = numpy.tile(numpy.append(initial_dipole, initial_position), self.n)
		if use_root and self.m == 6 * self.n:
			# We are perfectly specified
			return scipy.optimize.root(self.costs(), initial, jac=self.jac(), tol=1e-12)
		return scipy.optimize.least_squares(self.costs(), initial, jac=self.jac(), ftol=1e-15, gtol=3e-16)
