from dataclasses import dataclass
import numpy
import numpy.typing
from typing import Sequence, List
from pathfinder.model.static.dot import DotMeasurement


@dataclass
class StaticDipole():
	'''
	Representation of a static dipoles, either known or guessed.

	Parameters
	----------
	p : numpy.ndarray
		The static dipole moment.
	s : numpy.ndarray
		The position of the dipole.
	'''
	p: numpy.ndarray
	s: numpy.ndarray

	def __post_init__(self) -> None:
		'''
		Coerce the inputs into numpy arrays.
		'''
		self.p = numpy.array(self.p)
		self.s = numpy.array(self.s)

	def v_at_position(self, r: numpy.ndarray) -> float:
		'''
		Mirrors the same function in pathfinder.model.dot.
		'''
		diff = r - self.s
		return self.p.dot(diff) / (numpy.linalg.norm(diff)**3)


class StaticDipoleArrangement():
	'''
	A collection of static dipoles, which we are interested in being able to characterise.

	Parameters
	--------
	dipoles : Sequence[StaticDipole]
	'''
	def __init__(self, dipoles):
		self.dipoles = dipoles

	def get_dot_measurement(self, dot_position: numpy.typing.ArrayLike) -> DotMeasurement:
		r = numpy.array(dot_position)
		return DotMeasurement(sum([dipole.v_at_position(r) for dipole in self.dipoles]), r)

	def get_dot_measurements(self, dot_positions: Sequence[numpy.typing.ArrayLike]) -> List[DotMeasurement]:
		'''
		For a series of points, each with three coordinates, return
		'''

		return [self.get_dot_measurement(dot_position) for dot_position in dot_positions]
